
/* module.c: Sony PH Key Driver.
 *
 * Copyright 2017 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>
#include <linux/iio/consumer.h>
#include <linux/iio/types.h>
#include <asm/uaccess.h>
#include <icx_key.h>
#include "ph_key.h"


#define DRV_NAME      "ph_key"

/* #define DBUG_LOG */
#if defined(DBUG_LOG)
	#define DBGTRACE(fmt, ...)            printk(KERN_DEBUG "\e[32m[" DRV_NAME "] "       fmt "\e[0m\n", ## __VA_ARGS__)
#else
	#define DBGTRACE(fmt, ...)
#endif
#define ERRTRACE(fmt, ...)                printk(KERN_ERR "[" DRV_NAME "] [ERROR] \n" fmt, ## __VA_ARGS__)


/* Device tree */
#define PROP_AD_KEY_NAME      ("ad_key")
#define PROP_GPIO_KEY_NAME    ("gpio_key")
#define PROP_VIRTUAL_KEY_NAME ("virtual_key")
#define PROP_KEY_DATA_NAME    ("key_data")
#define PROP_KEY_FLAG_NAME    ("key_flag")
#define PROP_KEY_PIN_NAME     ("key_pin")
#define PROP_IO_CH_NAME       ("io-channel-name")
#define PROP_INT_NAME         ("interrupt-name")

#define for_each_child_of_node_by_name(from, name, dn) \
	for (dn = of_find_node_by_name(from, name); dn; \
	     dn = of_find_node_by_name(dn, name))


#define ADKEY_MAX             8

enum {
	SCAN_INTERVAL_CHAT_GUARD    =  10,
	SCAN_INTERVAL_UNPUSH_DETECT =  50,
	SCAN_INTERVAL_POLLING       = 100,
	SCAN_WAKEUP_DELAY           = 500,
	ADKEY_CHAT_TRIAL_CNT        =   4,
	GPIOKEY_CHAT_TRIAL_CNT      =  10
};


/* Key count */
struct key_cnt {
	int num;
	int count;
};

/* Key count info */
struct key_cnt_info {
	struct key_cnt akey;
	struct key_cnt gkey;
	struct key_cnt vkey;
};

/* AD key info */
struct phkey_akeyinfo {
	int                          flags;          /* Flag                  */
	int                          eint_num;       /* EINT number           */
	const char                  *channel;        /* Channel name          */
	int                          count;          /* AD key data count     */
	struct phkey_akeydata const *p_data;         /* AD key data           */
};

/* EINT */
struct eint {
	int                         flags;           /* Flag                  */
	int                         num;             /* EINT number           */
	const char                 *name;            /* EINT name             */
	u8                          eint_enable;     /* EINT enable flag      */
};

/* AD key */
struct akey {
	int                         state;           /* Current status        */
	int                         state_old[2];    /* Old status            */
	struct phkey_akeyinfo       info;            /* AD key info           */
	struct phkey_akeydata       data[ADKEY_MAX]; /* AD key data           */
	struct iio_channel         *chan;            /* AD channel            */
};

/* GPIO key */
struct gkey {
	int                         eint_num;        /* EINT number           */
	u16                         neg;             /* Flip flag             */
	u16                         state;           /* Current status        */
	u16                         state_old[2];    /* Old status            */
	int                         port;            /* Port number           */
	int                         key_code;        /* Key code              */
	bool                        wakeup_flag;     /* Wake up flag          */
};

/* Private data */
struct priv {
	struct miscdevice           misc;            /* MISC device           */
	struct input_dev           *p_input;         /* Input device          */
	u8                          suspend;         /* Suspend status flag   */
	u16                         key_count;       /* Key code count        */
	u32                        *p_key_code;      /* Key code list         */

	struct {
		struct delayed_work     work;            /* Delay work            */
		u8                      count;           /* Key count             */
		u8                      eint_count;      /* EINT count            */
		u8                      polling;         /* Polling flag          */
		u8                      dummy[1];        /* Dummy                 */
		struct akey            *p_key;           /* AD key                */
		struct eint            *p_eint;          /* EINT                  */
	} akey;

	struct {
		struct delayed_work     work;            /* Delay work            */
		u8                      count;           /* Key count             */
		u8                      eint_count;      /* EINT count            */
		u8                      polling;         /* Polling flag          */
		u8                      dummy[1];        /* Dummy                 */
		struct gkey            *p_key;           /* GPIO Key              */
		struct eint            *p_eint;          /* EINT                  */
	} gkey;
};

static struct priv             *m_p_priv;


/* -------------------------------------------------------------------------------- */
/*   Calc key count info                                                            */
/*                                                                                  */
/*   IN : p_node = device node                                                      */
/*        p_cnt  = key count info                                                   */
/*   OUT: 0                                                                         */
/*        -1    == Error                                                            */
/* -------------------------------------------------------------------------------- */
static int ph_key_get_key_count(struct device_node *p_node, struct key_cnt_info *p_cnt)
{
	int                    keydata_size = 0;
	int                    akey_count   = 0;
	int const             *p_keydata    = NULL;
	struct device_node    *p_child      = NULL;

	for_each_available_child_of_node(p_node, p_child) {
		if (!strcmp(p_child->name, PROP_AD_KEY_NAME)) {
			p_keydata = of_get_property(p_child, PROP_KEY_DATA_NAME, &keydata_size);
			if (p_keydata) {
				p_cnt->akey.num++;
				akey_count = keydata_size / sizeof(struct phkey_akeydata);
				if (akey_count <= ADKEY_MAX) {
					p_cnt->akey.count += akey_count;
				} else {
					ERRTRACE("%s(%d) : akey.count[%d]", __func__, __LINE__, p_cnt->akey.count);
					return -1;
				}
			}
		} else if (!strcmp(p_child->name, PROP_GPIO_KEY_NAME)) {
			p_keydata = of_get_property(p_child, PROP_KEY_DATA_NAME, &keydata_size);
			if (p_keydata) {
				p_cnt->gkey.num++;
				p_cnt->gkey.count += keydata_size / sizeof(int);
			}
		} else if (!strcmp(p_child->name, PROP_VIRTUAL_KEY_NAME)) {
			p_keydata  = of_get_property(p_child, PROP_KEY_DATA_NAME, &keydata_size);
			if (p_keydata) {
				p_cnt->vkey.num++;
				p_cnt->vkey.count += keydata_size / sizeof(int);
			}
		} else {
			/* nop */
		}
	}

	DBGTRACE("akey.num[%d]  akey.count[%d]", p_cnt->akey.num, p_cnt->akey.count);
	DBGTRACE("gkey.num[%d]  gkey.count[%d]", p_cnt->gkey.num, p_cnt->gkey.count);
	DBGTRACE("vkey.num[%d]  vkey.count[%d]", p_cnt->vkey.num, p_cnt->vkey.count);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   EINT Setting                                                                   */
/*                                                                                  */
/*   IN : p_eint = EINT address                                                     */
/*        count  = EINT count                                                       */
/*        num    = EINT number                                                      */
/*        name   = EINT name                                                        */
/*   OUT: EINT count                                                                */
/* -------------------------------------------------------------------------------- */
static inline u8 set_eint(struct eint *p_eint, u8 count, int flags, int num, const char *name)
{
	uint index = 0;

	for (index = 0; index < count; index++) {
		if (p_eint[index].num == num) {
			return count;
		}
	}

	p_eint[count].flags       = flags;
	p_eint[count].num         = num;
	p_eint[count].name        = name;
	p_eint[count].eint_enable = 1;

	return (count+1);
}


/* -------------------------------------------------------------------------------- */
/*   EINT mask setting                                                              */
/*                                                                                  */
/*   IN : enable  = 0:Disable, 1:Enable                                             */
/*        polling = Polling flag                                                    */
/*        p_eint  = EINT                                                            */
/*        count   = EINT count                                                      */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void mask_eint(uint enable, uint polling, struct eint *p_eint, uint count)
{
	struct eint *p_eint_end = NULL;
	uint         unmask     = 0;

	p_eint_end = &p_eint[count];

	while (p_eint != p_eint_end) {
		if (polling) {
			unmask = (!enable && (p_eint->flags & PHKEY_FLAG_WAKEUP));
		} else {
			unmask = (enable || (p_eint->flags & PHKEY_FLAG_WAKEUP));
		}

		if (p_eint->eint_enable != unmask) {
			if (unmask) {
				DBGTRACE("unmask = %d", p_eint->num);
				enable_irq(p_eint->num);
			} else {
				DBGTRACE("mask = %d", p_eint->num);
				disable_irq(p_eint->num);
			}
			p_eint->eint_enable = unmask;
		}

		p_eint++;
	}

	return;
}


/* -------------------------------------------------------------------------------- */
/*   EINT all mask setting                                                          */
/*                                                                                  */
/*   IN : p_priv = Private data                                                     */
/*        enable = 0:Disable, 1:Enable                                              */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void enable_eint(struct priv *p_priv, uint enable)
{
	mask_eint(enable, p_priv->akey.polling, &p_priv->akey.p_eint[0], p_priv->akey.eint_count);
	mask_eint(enable, p_priv->gkey.polling, &p_priv->gkey.p_eint[0], p_priv->gkey.eint_count);

	return;
}


/* -------------------------------------------------------------------------------- */
/*   AD key info setting                                                            */
/*                                                                                  */
/*   IN : p_priv = private data                                                     */
/*        p_node = device node                                                      */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_akey_info(struct priv *p_priv, struct device_node *p_node)
{
	uint        index        = 0;
	uint        count        = 0;
	uint        eint_count   = 0;
	int         keydata_size = 0;
	int         result       = 0;
	const char *eint_name    = DRV_NAME;

	struct device_node          *p_child = NULL;
	struct phkey_akeydata const *p_data  = NULL;

	for_each_child_of_node_by_name(p_node, PROP_AD_KEY_NAME, p_child) {
		result = of_property_read_u32(p_child, PROP_KEY_FLAG_NAME, &(p_priv->akey.p_key[index].info.flags));
		p_data = of_get_property(p_child, PROP_KEY_DATA_NAME, &keydata_size);
		if (!result && p_data) {
			of_property_read_u32_array(p_child, PROP_KEY_DATA_NAME, (u32 *)&p_priv->akey.p_key[index].info.p_data[0], (keydata_size / sizeof(int)));
			result = of_property_read_string(p_child, PROP_IO_CH_NAME, &(p_priv->akey.p_key[index].info.channel));
			if (!result) {
				p_priv->akey.p_key[index].info.count = keydata_size / sizeof(struct phkey_akeydata);
				p_priv->akey.p_key[index].state      = p_priv->akey.p_key[index].info.count;

				if (!(p_priv->akey.p_key[index].info.flags & PHKEY_FLAG_POLLING)) {
					of_property_read_string(p_child, PROP_INT_NAME, &eint_name);

					p_priv->akey.p_key[index].info.eint_num = irq_of_parse_and_map(p_child, 0);
					p_priv->akey.eint_count                 = set_eint(&p_priv->akey.p_eint[0],
																		p_priv->akey.eint_count,
																		p_priv->akey.p_key[index].info.flags,
																		p_priv->akey.p_key[index].info.eint_num,
																		eint_name);
					eint_count++;
				}

				for (count = 0; count < p_priv->akey.p_key[index].info.count; count++) {
					p_priv->p_key_code[p_priv->key_count++] = p_priv->akey.p_key[index].info.p_data[count].key_code;
					__set_bit(p_priv->akey.p_key[index].info.p_data[count].key_code, p_priv->p_input->keybit);
				}

				index++;
			} else {
				ERRTRACE("%s(%d) : result[%d]", __func__, __LINE__, result);
			}
		} else {
			ERRTRACE("%s(%d) : result[%d], p_data[%s]", __func__, __LINE__, result, (p_data)?"Not NULL":"NULL");
		}
	}

	p_priv->akey.polling = (eint_count != p_priv->akey.count);

	return;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO key info setting                                                          */
/*                                                                                  */
/*   IN : p_priv = private data                                                     */
/*        p_node = device node                                                      */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_gkey_info(struct priv *p_priv, struct device_node *p_node)
{
	uint                index      = 0;
	uint                eint_count = 0;
	int                 key_flags  = 0;
	int                 result     = 0;
	enum of_gpio_flags  gpio_flags = 0;
	struct device_node *p_child    = NULL;
	const char         *eint_name  = DRV_NAME;

	for_each_child_of_node_by_name(p_node, PROP_GPIO_KEY_NAME, p_child) {
		result = of_property_read_u32(p_child, PROP_KEY_FLAG_NAME, &key_flags);
		p_priv->gkey.p_key[index].port = of_get_named_gpio_flags(p_child, PROP_KEY_PIN_NAME, 0, &gpio_flags);
		if (!result && (p_priv->gkey.p_key[index].port >= 0)) {
			if (gpio_flags & OF_GPIO_ACTIVE_LOW) {
				p_priv->gkey.p_key[index].neg = 1;
			}

			result = of_property_read_u32(p_child, PROP_KEY_DATA_NAME, &(p_priv->gkey.p_key[index].key_code));
			if (!result) {
				result = of_property_read_string(p_child, PROP_INT_NAME, &eint_name);
				if (!result) {
					if (!(key_flags & PHKEY_FLAG_POLLING)) {
						p_priv->gkey.p_key[index].eint_num = irq_of_parse_and_map(p_child, 0);
						p_priv->gkey.eint_count            = set_eint(&p_priv->gkey.p_eint[0],
																		p_priv->gkey.eint_count,
																		key_flags,
																		p_priv->gkey.p_key[index].eint_num,
																		eint_name);
						eint_count++;
					}

					p_priv->gkey.p_key[index].wakeup_flag = false;

					gpio_request(p_priv->gkey.p_key[index].port, eint_name);

					p_priv->p_key_code[p_priv->key_count++] = p_priv->gkey.p_key[index].key_code;
					__set_bit(p_priv->gkey.p_key[index].key_code, p_priv->p_input->keybit);

					index++;
				} else {
					ERRTRACE("%s(%d) : result[%d]", __func__, __LINE__, result);
				}
			} else {
				ERRTRACE("%s(%d) : result[%d]", __func__, __LINE__, result);
			}
		} else {
			ERRTRACE("%s(%d) : result[%d], port[%d]", __func__, __LINE__, result, p_priv->gkey.p_key[index].port);
		}
	}

	p_priv->gkey.polling  = (eint_count != p_priv->gkey.count);

	return;
}


/* -------------------------------------------------------------------------------- */
/*   Update AD key state                                                            */
/*                                                                                  */
/*   IN : p_priv = Private data                                                     */
/*   OUT: -1    != Update count                                                     */
/*        -1    == Error                                                            */
/* -------------------------------------------------------------------------------- */
static int update_akey_state(struct priv *p_priv)
{
	int                          result      = 0;
	int                          index       = 0;
	int                          count       = 0;
	int                          value       = 0;
	struct akey                 *p_akey      = &p_priv->akey.p_key[0];
	struct akey                 *p_akey_end  = &p_priv->akey.p_key[p_priv->akey.count];
	struct phkey_akeydata const *p_akey_data = NULL;
	struct phkey_akeyinfo const *p_akey_info = NULL;

	do {
		p_akey_info          = &(p_akey->info);
		p_akey->state_old[1] = p_akey->state_old[0];
		result               = iio_read_channel_processed(p_akey->chan, &value);
		if (result < 0) {
			return -1;
		}

		p_akey_data = p_akey_info->p_data;
		for (index = 0; index < p_akey_info->count; index++, p_akey_data++) {
			if ((u16)(value - p_akey_data->min) <= (p_akey_data->max - p_akey_data->min)) {
				break;
			}
		}
		p_akey->state_old[0] = index;

		count += (p_akey->state_old[0] == p_akey->state_old[1]);
	} while (++p_akey != p_akey_end);

	return count;
}


/* -------------------------------------------------------------------------------- */
/*   Update AD key event                                                            */
/*                                                                                  */
/*   IN : p_priv = Private data                                                     */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void update_akey_event(struct priv *p_priv)
{
	int                          count       = 0;
	struct akey                 *p_akey      = &p_priv->akey.p_key[0];
	struct akey                 *p_akey_end  = &p_akey[p_priv->akey.count];
	struct phkey_akeyinfo const *p_akey_info = NULL;

	do {
		p_akey_info = &(p_akey->info);

		if (p_akey->state != p_akey->state_old[0]) {
			if (p_akey->state != p_akey_info->count) {
				input_event(p_priv->p_input, EV_KEY, p_akey_info->p_data[p_akey->state].key_code, 0);
				input_sync(p_priv->p_input);
			}

			p_akey->state = p_akey->state_old[0];
			if (p_akey->state != p_akey_info->count) {
				input_event(p_priv->p_input, EV_KEY, p_akey_info->p_data[p_akey->state].key_code, 1);
				input_sync(p_priv->p_input);
			}
		}

		count += (p_akey->state != p_akey_info->count);
	} while (++p_akey != p_akey_end);

	if (count) {                      /* Pushing */
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_INTERVAL_UNPUSH_DETECT));
	} else {
		if (p_priv->akey.polling) {   /* Polling */
			schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));
		} else {
			DBGTRACE("ad - stop");
		}
	}

	return;
}


/* -------------------------------------------------------------------------------- */
/*   Update GPIO key state                                                          */
/*                                                                                  */
/*   IN : p_priv = Private data                                                     */
/*   OUT: Non update count                                                          */
/* -------------------------------------------------------------------------------- */
static int update_gkey_state(struct priv *p_priv)
{
	int          count      = 0;
	struct gkey *p_gkey     = &p_priv->gkey.p_key[0];
	struct gkey *p_gkey_end = &p_gkey[p_priv->gkey.count];

	do {
		p_gkey->state_old[1] = p_gkey->state_old[0];
		p_gkey->state_old[0] = gpio_get_value(p_gkey->port) ^ p_gkey->neg;

		count += (p_gkey->state_old[0] == p_gkey->state_old[1]);
	} while (++p_gkey != p_gkey_end);

	return count;
}


/* -------------------------------------------------------------------------------- */
/*   Update GPIO key event                                                          */
/*                                                                                  */
/*   IN : p_priv = Private data                                                     */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void update_gkey_event(struct priv *p_priv)
{
	int          count      = 0;
	struct gkey   *p_gkey     = &p_priv->gkey.p_key[0];
	struct gkey   *p_gkey_end = &p_gkey[p_priv->gkey.count];

	do {
		if (p_gkey->state != p_gkey->state_old[0]) {
			p_gkey->state  = p_gkey->state_old[0];
			input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, p_gkey->state);
			input_sync(p_priv->p_input);
		}

		count += p_gkey->state;
	} while (++p_gkey != p_gkey_end);

	if (count) {                      /* Pushing */
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies(SCAN_INTERVAL_UNPUSH_DETECT));
	} else {
		if (p_priv->gkey.polling) {   /* Polling */
			schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));
		} else {
			DBGTRACE("gpio - stop");
		}
	}

	return;
}


/* -------------------------------------------------------------------------------- */
/*   AD key interrupt handler                                                       */
/*                                                                                  */
/*   IN : irq  = IRQ number                                                         */
/*        data = data                                                               */
/*   OUT: status                                                                    */
/* -------------------------------------------------------------------------------- */
static irqreturn_t akey_isr(int irq, void *data)
{
	struct priv *p_priv = (struct priv *)data;

	DBGTRACE("akey_isr");

	if (!p_priv->suspend) {
		schedule_delayed_work(&p_priv->akey.work, 0);
	}

	return IRQ_HANDLED;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO key interrupt handler                                                     */
/*                                                                                  */
/*   IN : irq  = IRQ number                                                         */
/*        data = data                                                               */
/*   OUT: status                                                                    */
/* -------------------------------------------------------------------------------- */
static irqreturn_t gkey_isr(int irq, void *data)
{
	struct priv *p_priv = (struct priv *)data;
	struct gkey *p_gkey = NULL;

	DBGTRACE("gkey_isr");

	if (!p_priv->suspend) {
		schedule_delayed_work(&p_priv->gkey.work, 0);
	} else {
		p_gkey = &p_priv->gkey.p_key[0];
		p_gkey->wakeup_flag = true;
	}

	return IRQ_HANDLED;
}


/* -------------------------------------------------------------------------------- */
/*   AD key work queue process                                                      */
/*                                                                                  */
/*   IN : p_work = work                                                             */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void akey_workqueue_proc(struct work_struct *p_work)
{
	struct priv *p_priv = container_of(p_work, struct priv, akey.work.work);
	int          result = 0;
	int          retry  = 0;

	DBGTRACE("ad - start");

	result = update_akey_state(p_priv);
	if (result < 0) {
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));
		return;
	}

	for (retry = 1; retry < ADKEY_CHAT_TRIAL_CNT; retry++) {
		msleep_interruptible(SCAN_INTERVAL_CHAT_GUARD);

		result = update_akey_state(p_priv);
		if (result < 0) {
			break;
		}

		if (result == p_priv->akey.count) {
			update_akey_event(p_priv);
			return;
		}
	}

	schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));

	return;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO key work queue process                                                    */
/*                                                                                  */
/*   IN : p_work = work                                                             */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void gkey_workqueue_proc(struct work_struct *p_work)
{
	struct priv *p_priv = container_of(p_work, struct priv, gkey.work.work);
	int          result = 0;
	int          retry  = 0;

	DBGTRACE("gpio - start");

	update_gkey_state(p_priv);

	for (retry = 1; retry < GPIOKEY_CHAT_TRIAL_CNT; retry++) {
		result = update_gkey_state(p_priv);

		if (result == p_priv->gkey.count) {
			update_gkey_event(p_priv);
			return;
		}
	}

	schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));

	return;
}


/* -------------------------------------------------------------------------------- */
/*   Open                                                                           */
/*                                                                                  */
/*   IN : p_inode = i node                                                          */
/*        p_file  = file                                                            */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, struct priv, misc);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*  Write                                                                           */
/*                                                                                  */
/*  IN : p_file = file                                                              */
/*       p_buff = write buffer                                                      */
/*       size   = write size                                                        */
/*       p_pos  = possition                                                         */
/*  OUT: write size                                                                 */
/* -------------------------------------------------------------------------------- */
static ssize_t ph_key_write(struct file *p_file, const char __user *p_buff, size_t size, loff_t *p_pos)
{
	struct priv *p_priv   = p_file->private_data;
	char         buff[32] = {0};
	char        *p_end    = NULL;
	u32          index    = 0;
	u32          value    = 0;

	if (size >= sizeof(buff)) {
		return -EINVAL;
	}

	if (copy_from_user(&buff[0], &p_buff[0], size)) {
		return -EFAULT;
	}

	/* Index */
	index = simple_strtoul(&buff[0], &p_end, 0);
	if ((index >= p_priv->key_count) || (*p_end != ' ')) {
		return -EINVAL;
	}

	/* Setting value */
	value = simple_strtoul(&p_end[1], NULL, 0);
	if (value > 1) {
		return -EINVAL;
	}

	input_event(p_priv->p_input, EV_KEY, p_priv->p_key_code[index], value);
	input_sync(p_priv->p_input);

	return size;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = file                                                             */
/*        cmd    = command                                                          */
/*        arg    = argument                                                         */
/*   OUT: 0                                                                         */
/*        key count                                                                 */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long ph_key_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct priv *p_priv = p_file->private_data;

	switch (cmd) {
	case ICX_KEY_IOCTL_CMD_GETCOUNT:
		{
			return p_priv->key_count;
		}

	case ICX_KEY_IOCTL_CMD_GETINFO:
		{
			if (copy_to_user((void __user *)arg, &p_priv->p_key_code[0], sizeof(u32)*p_priv->key_count)) {
				return -EFAULT;
			}
			break;
		}

	default:
		return -EINVAL;
	}

	return 0;
}


static const struct file_operations ph_key_fops = {
										.owner          = THIS_MODULE,
										.open           = ph_key_open,
										.write          = ph_key_write,
										.unlocked_ioctl = ph_key_ioctl,
										.compat_ioctl   = ph_key_ioctl,
									};


/* -------------------------------------------------------------------------------- */
/*   Probe                                                                          */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/*        -ENOMEM                                                                   */
/* -------------------------------------------------------------------------------- */
static int ph_key_probe(struct platform_device *p_dev)
{
	int                  result      = 0;
	uint                 index       = 0;
	uint                 count       = 0;
	uint                 size        = 0;
	int                  irq_trigger = IRQF_TRIGGER_NONE;
	struct priv         *p_priv      = NULL;
	struct eint         *p_eint      = NULL;
	struct device_node  *p_node      = NULL;
	struct key_cnt_info  cnt_info    = { {0, 0}, {0, 0}, {0, 0} };

	/* Get key code count */
	result = ph_key_get_key_count(p_dev->dev.of_node, &cnt_info);
	if (result) {
		ERRTRACE("%s(%d)", __func__, __LINE__);
		return -ENOMEM;
	}
	count = cnt_info.akey.count + cnt_info.gkey.count + cnt_info.vkey.count;

	/* Allcate private data */
	size = sizeof(struct priv) +
			((sizeof(struct akey) + sizeof(struct eint)) * cnt_info.akey.num) +
			(sizeof(struct phkey_akeyinfo) * cnt_info.akey.num) +
			(sizeof(struct phkey_akeydata) * cnt_info.akey.count) +
			((sizeof(struct gkey) + sizeof(struct eint)) * cnt_info.gkey.num) +
			(sizeof(u32) * count);
	p_priv = kzalloc(size, GFP_KERNEL);
	if (p_priv == NULL) {
		ERRTRACE("%s(%d)", __func__, __LINE__);
		return -ENOMEM;
	}

	p_priv->akey.count  = cnt_info.akey.num;
	p_priv->gkey.count  = cnt_info.gkey.num;
	p_priv->akey.p_key  = (struct akey *)&p_priv[1];
	p_priv->gkey.p_key  = (struct gkey *)&p_priv->akey.p_key[p_priv->akey.count];
	p_priv->akey.p_eint = (struct eint *)&p_priv->gkey.p_key[p_priv->gkey.count];
	p_priv->gkey.p_eint = (struct eint *)&p_priv->akey.p_eint[p_priv->akey.count];
	p_priv->p_key_code  = (u32         *)&p_priv->gkey.p_eint[p_priv->gkey.count];
	p_priv->key_count   = 0;

	for (index = 0; index < cnt_info.akey.num; index++) {
		p_priv->akey.p_key[index].info.p_data = &(p_priv->akey.p_key[index].data[0]);
	}

	INIT_DELAYED_WORK(&p_priv->akey.work, akey_workqueue_proc);
	INIT_DELAYED_WORK(&p_priv->gkey.work, gkey_workqueue_proc);

	/* Create MISC device */
	p_priv->misc.minor = MISC_DYNAMIC_MINOR;
	p_priv->misc.name  = DRV_NAME;
	p_priv->misc.fops  = &ph_key_fops;
	result             = misc_register(&p_priv->misc);
	if (result) {
		ERRTRACE("%s(%d)", __func__, __LINE__);
		goto ph_key_probe_f00;
	}

	/* Allcate input device */
	p_priv->p_input = input_allocate_device();
	if (p_priv->p_input == NULL) {
		result = -ENOMEM;
		ERRTRACE("%s(%d)", __func__, __LINE__);
		goto ph_key_probe_e00;
	}

	/* Set virtual key */
	p_node = of_find_node_by_name(p_dev->dev.of_node, PROP_VIRTUAL_KEY_NAME);
	if (!p_node) {
		ERRTRACE("%s(%d)", __func__, __LINE__);
		goto ph_key_probe_e00;
	}
	result = of_property_read_u32_array(p_node, PROP_KEY_DATA_NAME, &p_priv->p_key_code[0], cnt_info.vkey.count);
	if (result < 0) {
		ERRTRACE("%s(%d) : result[%d]", __func__, __LINE__, result);
	} else {
		for (index = 0; index < cnt_info.vkey.count; index++) {
			__set_bit(p_priv->p_key_code[index], p_priv->p_input->keybit);
			p_priv->key_count++;
		}
	}

	/* Set AD key */
	set_akey_info(p_priv, p_dev->dev.of_node);

	/* Set GPIO key */
	set_gkey_info(p_priv, p_dev->dev.of_node);

	/* Register input device */
	__set_bit(EV_KEY, p_priv->p_input->evbit);
	p_priv->p_input->name       = DRV_NAME;
	p_priv->p_input->phys       = DRV_NAME;
	p_priv->p_input->id.bustype = BUS_HOST;
	p_priv->p_input->id.vendor  = 0x0001;
	p_priv->p_input->id.product = 0x0001;
	p_priv->p_input->id.version = 0x0100;
	p_priv->p_input->dev.parent = &p_dev->dev;

	result = input_register_device(p_priv->p_input);
	if (result) {
		ERRTRACE("%s(%d)", __func__, __LINE__);
		goto ph_key_probe_d00;
	}

	platform_set_drvdata(p_dev, p_priv);
	m_p_priv = p_priv;

	if (p_priv->akey.polling) {
		/* AD key polling setting */
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));
	} else {
		/* AD key interrupt setting */
		p_eint = &p_priv->akey.p_eint[0];
		for (index = 0; index < p_priv->akey.eint_count; index++, p_eint++) {
			if (p_eint->flags & PHKEY_FLAG_HIGH) {
				irq_trigger = IRQF_TRIGGER_RISING;
			} else {
				irq_trigger = IRQF_TRIGGER_FALLING;
			}
			result = request_irq(p_eint->num, akey_isr, irq_trigger, p_eint->name, p_priv);

			if (p_eint->flags & PHKEY_FLAG_WAKEUP) {
				irq_set_irq_wake(p_eint->num, 1);
			}
		}
	}

	/* Get ADC channel */
	for (index = 0; index < p_priv->akey.count; index++) {
		p_priv->akey.p_key[index].chan = iio_channel_get(&p_dev->dev, p_priv->akey.p_key[index].info.channel);
	}

	if (p_priv->gkey.polling) {
		/* GPIO key polling setting */
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies(SCAN_INTERVAL_POLLING));
	} else {
		/* GPIO key interrupt setting */
		p_eint = &p_priv->gkey.p_eint[0];
		for (index = 0; index < p_priv->gkey.eint_count; index++, p_eint++) {
			if (p_eint->flags & PHKEY_FLAG_HIGH) {
				irq_trigger = IRQF_TRIGGER_RISING;
			} else {
				irq_trigger = IRQF_TRIGGER_FALLING;
			}
			result = request_irq(p_eint->num, gkey_isr, irq_trigger, p_eint->name, p_priv);

			if (p_eint->flags & PHKEY_FLAG_WAKEUP) {
				irq_set_irq_wake(p_eint->num, 1);
			}
		}
	}

	return 0;

ph_key_probe_d00:
	input_free_device(p_priv->p_input);
ph_key_probe_e00:
	misc_deregister(&p_priv->misc);
ph_key_probe_f00:
	kfree(p_priv);

	return result;
}


/* -------------------------------------------------------------------------------- */
/*   Remove                                                                         */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_remove(struct platform_device *p_dev)
{
	struct priv *p_priv      = platform_get_drvdata(p_dev);
	struct eint *p_eint      = NULL;
	struct gkey *p_gkey      = NULL;
	struct akey *p_akey      = &p_priv->akey.p_key[0];
	struct akey *p_akey_end  = &p_akey[p_priv->akey.count];
	uint         index       = 0;

	/* Free IRQ */
	p_eint = &p_priv->akey.p_eint[0];
	for (index = 0; index < p_priv->akey.eint_count; index++, p_eint++) {
		if (p_eint->flags & PHKEY_FLAG_WAKEUP) {
			irq_set_irq_wake(p_eint->num, 0);
		}
		free_irq(p_eint->num, p_priv);
	}

	p_eint = &p_priv->gkey.p_eint[0];
	for (index = 0; index < p_priv->gkey.eint_count; index++, p_eint++) {
		if (p_eint->flags & PHKEY_FLAG_WAKEUP) {
			irq_set_irq_wake(p_eint->num, 0);
		}
		free_irq(p_eint->num, p_priv);
	}

	/* Free GPIO */
	p_gkey = &p_priv->gkey.p_key[0];
	for (index = 0; index < p_priv->gkey.count; index++, p_gkey++) {
		gpio_free(p_gkey->port);
	}

	/* Free ADC */
	do {
		iio_channel_release(p_akey->chan);
	} while (++p_akey != p_akey_end);

	/* Cancel delay work queue */
	cancel_delayed_work_sync(&p_priv->akey.work);
	cancel_delayed_work_sync(&p_priv->gkey.work);

	input_unregister_device(p_priv->p_input);

	misc_deregister(&p_priv->misc);

	kfree(p_priv);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Suspend                                                                        */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*        state = status                                                            */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_suspend(struct platform_device *p_dev, pm_message_t state)
{
	struct priv *p_priv = platform_get_drvdata(p_dev);
	struct gkey *p_gkey = NULL;
	uint         index  = 0;

	DBGTRACE("suspend");

	p_priv->suspend = 1;

	enable_eint(p_priv, 0);

	if (p_priv->akey.count) {
		cancel_delayed_work_sync(&p_priv->akey.work);
	}

	if (p_priv->gkey.count) {
		cancel_delayed_work_sync(&p_priv->gkey.work);
	}

	/* Force release press key */
	p_gkey = &p_priv->gkey.p_key[0];
	for (index = 0; index < p_priv->gkey.count; index++, p_gkey++) {
		if (p_gkey->state) {
			input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, 0);
			input_sync(p_priv->p_input);
			p_gkey->state = 0;
		}
		p_gkey->state_old[0] = 0;
		p_gkey->state_old[1] = 0;
	}

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Resume                                                                         */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_resume(struct platform_device *p_dev)
{
	struct priv *p_priv = platform_get_drvdata(p_dev);
	struct gkey *p_gkey = NULL;
	uint         index  = 0;

	DBGTRACE("resume");

	/* Detected Wake-up factor GPIO key press */
	if (!p_priv->gkey.polling) {
		p_gkey = &p_priv->gkey.p_key[0];
		for (index = 0; index < p_priv->gkey.count; index++, p_gkey++) {
			if (p_gkey->wakeup_flag) {
				p_gkey->state        = 1;
				p_gkey->state_old[0] = 1;
				p_gkey->state_old[1] = 1;
				input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, 1);
				input_sync(p_priv->p_input);
				p_gkey->wakeup_flag = false;
			}
		}
	}

	enable_eint(p_priv, 1);

	p_priv->suspend = 0;

	if (p_priv->akey.count) {
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies(SCAN_WAKEUP_DELAY));
	}

	if (p_priv->gkey.count) {
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies(SCAN_WAKEUP_DELAY));
	}

	return 0;
}


static const struct of_device_id ph_key_of_match[] = {
	{ .compatible = "sony,ph-key" },
	{ },
};

MODULE_DEVICE_TABLE(of, ph_key_of_match);


static struct platform_driver ph_key_platform_driver = {
									.probe   = ph_key_probe,
									.remove  = ph_key_remove,
									.suspend = ph_key_suspend,
									.resume  = ph_key_resume,
									.driver  = {
										.name           = DRV_NAME,
										.owner          = THIS_MODULE,
										.of_match_table = ph_key_of_match,
									},
								};


module_platform_driver(ph_key_platform_driver);


MODULE_DESCRIPTION("Sony PH Key Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_ALIAS("Sony PH Key Driver");
MODULE_VERSION("v2.00");

